library(XML)
.Files<-NULL
.Pack<-list(
	"Rendering/"= "R_VTK_RENDERING",
	"Widgets/"="R_VTK_WIDGETS",
	"VolumeRendering/"="R_VTK_RENDERING",
	"Common/"="R_VTK_BASE",
	"IO/"="R_VTK_BASE",
	"Imaging/"="R_VTK_BASE",
	"Filtering/"="R_VTK_BASE",
	"GenericFiltering/"="R_VTK_BASE",
	"Graphics/"="R_VTK_BASE",
	"Hybrid/"="R_VTK_HYBRID"
)
.VTK.ROOT <- "/usr/local/src/VTK/"
.SubLib <- "Common/"
.Nicks <- list(vtkIdType="long int")
.Hints <- read.table(paste(.VTK.ROOT,"Wrapping/hints",sep=""))
.RHints <- read.table("RHints")

.Find.Id <- function(tree,id){
  nodeFound <- NULL
  xmlApply(tree,function(x){
    if(xmlAttrs(x)[["id"]]==id){
      nodeFound <<- x
    }
  })
  return(nodeFound)
}

.Get.Func.Header <- function(node){
  fl <- xmlAttrs(node)[["file"]]
  if (is.null(.Files[[fl]])){
     .Files[[fl]]$handle <<- file(paste(.VTK.ROOT,.SubLib,xmlAttrs(.Find.Id(r,fl))[["name"]],sep=""))
     .Files[[fl]]$sline <<- 0
  }
  sline<-as.integer(xmlAttrs(node)[["line"]])
  z<-readLines(.Files[[fl]]$handle,sline-.Files[[fl]]$sline)

  .Files[[fl]]$sline<<-0
  grep("\\(",z)->x
  grep("\\)",z)->x1
  print(x1)
  print(x)
  z1<-paste(z[max(x):max(x1)],collapse="")
  return(z1)
}

.Get.Dim<-function(node,func.head){
  name <- xmlAttrs(node)[["name"]]
  dim <- 0
  z <- regexpr(paste(" ",name,"\\[[0-9]*\\]",sep=""),func.head,extended=T,perl=T)
  if (z[[1]]>0) {
    z1 <- substr(func.head,z[[1]]+1,z[[1]]+attributes(z)[[1]]-2)
    print(z1)
    z2 <- regexpr("\\[",z1,extended=T,perl=T)
    print(z2)
    dim <- as.integer(substr(z1,z2[[1]]+1,nchar(z1)))
    if (is.na(dim)) { dim<-0 }
    print(dim)
    z3 <- regexpr(paste(" ",name,"\\[[0-9]*\\]\\[[0-9]*\\]",sep=""),func.head,extended=T,perl=T)
    print(z3)
    if (z3[[1]]>0) {
       dim <- c(dim,as.integer(substr(func.head,z[[1]]+1+attributes(z)[[1]],z[[1]]+attributes(z3)[[1]]-2)))
       print(dim)
    }
  } else {
    #use hints
  }
  return(dim) 
}

.use.hints <- function(class,method){
  z<-(.Hints$V1==class)
  meths<-.Hints$V2[z]
  z1 <- (meths==method)
  h<-(.Hints$V4[z])[z1]
  if((length(h)==0)||is.null(h)||is.na(h)){
    h <- 1
  }
  return(h)
}

.check.rhints <- function(classname,methodname){
  for (i in seq(.RHints$V1)){
    if(length(grep(as.character(.RHints$V1[i]),classname))>0){
      if(length(grep(as.character(.RHints$V2[i]),methodname))>0){
        return(as.character(.RHints$V3[i]))
      }
    }   
  }
  return("continue")
}

convert.toRtype <- function(ctype){
  switch(ctype,
         int=return("as.integer"),
         "long int"=return("as.integer"),
         "long unsigned int"=return("as.integer"),
         double=return("as.double"),
         float=return("as.single"),
         char=return("as.character")
       )
}


.Process.Type <- function(node,typedesc){
  print(node)
  switch(xmlName(node),
         Method={
           nodeFound <- .Find.Id(tree=r,id=xmlAttrs(node)[["returns"]])
           typedesc <- .Process.Type(nodeFound,typedesc)
         },
         Argument={
           nodeFound <- .Find.Id(tree=r,id=xmlAttrs(node)[["type"]])
           typedesc <- .Process.Type(nodeFound,typedesc)
         },
         CvQualifiedType={
           nodeFound <- .Find.Id(tree=r,id=xmlAttrs(node)[["type"]])
           typedesc$const <- TRUE 
           typedesc <- .Process.Type(nodeFound,typedesc)
         },
         FundamentalType={
           name <- xmlAttrs(node)[["name"]]
           typedesc$type <- name
           typedesc$vtkclass <- F
         },
         Class={
           name <- xmlAttrs(node)[["name"]]
           typedesc$type <- name
           typedesc$vtkclass <- T
         },
         Struct={
           name <- xmlAttrs(node)[["name"]]
           if (regexpr("basic_.stream",name)>-1){
             stop(paste(name," -  Unhandled type"))
           }
           typedesc$type <- name
           typedesc$vtkclass <- T
         },
	 Ellipsis={
	   stop("Unsupported Argument")
	 },
         Typedef={
           nodeFound <- .Find.Id(tree=r,id=xmlAttrs(node)[["type"]])
           typedesc <- .Process.Type(nodeFound,typedesc)
         },
         FunctionType={
           typedesc$type<-"function"
         },
         ArrayType={
	   nodeFound <- .Find.Id(tree=r,id=xmlAttrs(node)[["type"]])
           typedesc <- .Process.Type(nodeFound,typedesc)
         },
         ReferenceType={
#           typedesc$reference <- 0
           typedesc$weak <- T
           nodeFound <- .Find.Id(tree=r,id=xmlAttrs(node)[["type"]])
           typedesc <- .Process.Type(nodeFound,typedesc)
         },
         PointerType={
           typedesc$reference <- typedesc$reference+1
           
           nodeFound <- .Find.Id(tree=r,id=xmlAttrs(node)[["type"]])
           typedesc <- .Process.Type(nodeFound,typedesc)
         })
  return(typedesc)       
}

.Parse.Class <- function(classname,filename=paste(.SubLib,classname,".xml",sep="")){
  xmlTreeParse(filename)->t
  .Has.New<-F
  r<<-xmlRoot(t)
  xmlApply(r,xmlName)->z
                                        #Find class with requested name
  r0 <- r[z=="Class"]
  which(unlist(lapply(r0,function(x){xmlAttrs(x)[["name"]]==classname})))->l
  xmlAttrs(r0[[l]])[["id"]]->z2
  xmlAttrs(r0[[l]])[["name"]]->classname
  !is.na(xmlAttrs(r0[[l]])["abstract"])->abstract.class
                                        #make cc preamble
  ccfile <- file(paste("../src/",classname,".cc",sep=""),"w+")
  cat("#include <",classname,".h>",file=ccfile,sep="",fill=T,append=T) #
  cat("extern \"C\" {",file=ccfile,sep="",fill=T,append=T)
  cat("#include <R.h>",file=ccfile,sep="",fill=T,append=T) #
  cat("#include <Rinternals.h>",file=ccfile,sep="",fill=T,append=T) #
  cat("#include <Rdefines.h>",file=ccfile,sep="",fill=T,append=T) #
  cat("#include \"Utils.h\"",file=ccfile,sep="",fill=T,append=T) #
  cat("#ifdef ",.Pack[[.SubLib]],file=ccfile,sep="",fill=T,append=T)
                                        #make R preamble
  Rfile <- file(paste("../R/",classname,".R",sep=""),"w+")
  xmlAttrs(r0[[l]])[["bases"]]->baseclassid
  print(baseclassid)
  which(unlist(xmlApply(r,function(x){match(paste(xmlAttrs(x)[["id"]]," ",sep=""),baseclassid)==1})))->bl
  print(bl)
  gsub("vtkDataArrayTemplate<.*>","vtkDataArray",xmlAttrs(r[[bl]])[["name"]])->baseclassname
  cat("if (",gsub("_",".",.Pack[[.SubLib]],fixed=T),"==1){",file=Rfile,sep="",fill=T,append=T)
  cat(classname," <- function(.me=environment()){",file=Rfile,sep="",fill=T,append=T)
  cat("  .me <- ",baseclassname,"(.me)",file=Rfile,sep="",fill=T,append=T)
                                        #find class methods
  r1 <- r[z=="Method"]
  which(unlist(lapply(r1,function(x){xmlAttrs(x)[["context"]]==z2})))->l2

                                        #process class methods
  .Method.Names <- NULL
  sapply(r1[l2],function(n){
                                        #check private-public?
    if (!is.na(match("access",names(xmlAttrs(n)))) ) {
                                        #private or protected - skipping
      return(F)
    }
  
                                        #get public method name
    methodname <- xmlAttrs(n)[["name"]]
    .Func.Header <- .Get.Func.Header(n)
                                        #Check if some special processing of function is required by looking into RHints file. Generally it is used to skip functions we in fact can not handle but can not guess this from its XML-representation.
    switch(.check.rhints(classname,methodname),
           omit={
             return(F)
           },
           hasnew={
             .Has.New<<-T
             return(F)
           },
           continue={
           }
           )
                                        #get arguments
    args <- xmlChildren(n)
    num.refs <- 0
    argdefs <- try(lapply(args,function(a){
      argtype <- .Process.Type(a,list(reference=0,weak=F,const=F,type="",vtkclass=F))
      if (argtype$reference>1){
        stop("Multidimensional arguments!")
      }
      argtype$name <- xmlAttrs(a)[["name"]]
      if (((argtype$weak)||(argtype$reference<1))&&(!((argtype$weak)&&(argtype$reference>0)))){
                                        #Some Sort of Black Magic. We consider the argument to have length 1 if it is either pointer or weak reference (but not both). Otherwise we look into source and try to get it length from grepping for argname[.*]. If this fails too then argument length is 0 and this means the VTK does not care about its length so it is up to end-user to provide R vector with right size. If in result the argument length is not 0 we later add a SET_LENGTH clause in .cc file to properly coerce its length.
        argtype$dim <- 1
      } else {
        argtype$dim <- .Get.Dim(a,.Func.Header)
      }
      if (length(argtype$dim)>1){
        stop("Multidimensional arguments!")
      }
                                        #count the number of reference args. All referenced arguments will be returned from .Call() function and will be stored in .VTK.Cache variable
      if (((argtype$weak)||(argtype$reference>0))&&(!argtype$const)) {
        num.refs <<- num.refs+1
      }
      return(argtype)
    }))
                                        #something wrong with arguments (or arguments is a multi-pointer) - invalid function
    if(class(argdefs)=="try-error"){
      return(F)
    }
    print(paste("argedefs=",length(argdefs)))

                                        #handle only methods which install callbacks with pointer to function like void * func( void *) and with second arguments to be passed to func as void parameter. Anything else we can not handle.  
    if ((length(argdefs)>0)&&(argdefs[[1]]$type=="function")){
      if (length(argdefs)!=2){
        return(F)
      } else {
        argdefs<-argdefs[1]
      }
    }
                                        #get return value type 
    retvaldef <- .Process.Type(n,list(reference=0,type="",dim=NULL,vtkclass=F))
                                        #There is no information in XML dump or in C-code about the length of return value if the value is a pointer. The last resort is to look into hints file provided by kitware programmers where the length of return value for the <methodname>-method of <classname>-class is written. If this fail we fallback to the safe  length 1.
    if (retvaldef$reference>0){
      retvaldef$dim <- .use.hints(classname,methodname)
    } else {
      retvaldef$dim <- 1
    }
    
                                        #append function name to R/cc files
    suffix <- ifelse((methcount <- sum(.Method.Names==methodname))==0,"",paste("_v",as.character(methcount),sep=""))
    cat("  .me$",methodname,suffix," <- function(",file=Rfile,sep="",fill=F,append=T)
    cat("  SEXP R_",classname,"_",methodname,suffix,"( SEXP obj", file=ccfile,sep="",fill=F,append=T) 
    start.head <- TRUE
    try(lapply(argdefs,function(aa){
      cat(", SEXP ",aa$name,sep="",file=ccfile,fill=F,append=T)
      if (start.head){
        start.head<<-FALSE
        cat(" .",aa$name,file=Rfile,sep="",fill=F,append=T)
        if (aa$type=="function"){
          cat(", SEXP env",sep="",file=ccfile,fill=F,append=T)
#          cat(", .env = environment()",file=Rfile,sep="",fill=F,append=T)
	 cat(", ...",file=Rfile,sep="",fill=F,append=T)
          stop("")
        }
      } else {
        cat(", .",aa$name,file=Rfile,sep="",fill=F,append=T)
      }
      if (((aa$weak)||(aa$reference>0))&&(!aa$const)){
        # Try to create fail-safe defaults to arguments which will be passed to C++-function as references. It is quite probable that in fact they are used as a side effect return values so theirs input values mean nothing. The code just save efforts of end-user he does not have to type into arguments which will never be used. 
        if (aa$vtkclass){
          cat("=",aa$type,"()",file=Rfile,sep="",fill=F,append=T)
        } else {
          if(prod(aa$dim)>0){
            cat("=",convert.toRtype(aa$type),"(rep.int(1,",prod(aa$dim),"))",file=Rfile,sep="",fill=F,append=T)
          }
        }
      }
    }))
    cat("){",file=Rfile,sep="",fill=T,append=T)
    cat("){",file=ccfile,sep="",fill=T,append=T)
                                        #temp cc-vars
    cat("    SEXP ans,res_names,res = NULL_USER_OBJECT;",file=ccfile,sep="",fill=T,append=T)

                                        #create c-vars devinitions
    try(lapply(argdefs,function(aa){
      if (aa$type=="function"){
        cat("     RVoidFuncArg * _arg=Calloc(1,RVoidFuncArg);" ,file=ccfile,sep="",fill=T,append=T)
        cat("     _arg->env=env;" ,file=ccfile,sep="",fill=T,append=T)
        cat("     _arg->func=",aa$name,";",file=ccfile,sep="",fill=T,append=T)
        cat("     ",paste(metodname,"ArgDelete(RVoidFuncArgDelete)",sep=""),file=ccfile,sep="",fill=T,append=T) 
        stop("")
      }
      if(prod(aa$dim)>0){
                                        #Set the length explicitly if we do know argument length
        cat("    SET_LENGTH(",aa$name,",",prod(aa$dim),");",sep="",file=ccfile,fill=T,append=T)
      }
                                        #convert SEXPs to c-type.
      cat("    ",aa$type," * ",paste("_",aa$name,sep=""),"=",ifelse(aa$vtkclass,paste("(",aa$type," * ",") ",sep=""),""),paste("SexpTo",ifelse(aa$vtkclass,"Vtk",gsub(" ","_",aa$type)),"(",aa$name,");",sep=""),file=ccfile,fill=T,append=T)

    }))
                                        #create obj definition
    cat("    ",classname," * _obj = (",classname, " * ) SexpToVtk(obj);",sep="",file=ccfile,fill=T,append=T)
    
                                        #create calls to vtk library function in cc file and call to wrapper library in R file
    cat("    ",ifelse(retvaldef$type=="void",ifelse(retvaldef$reference>0,"void",""),retvaldef$type),ifelse(retvaldef$reference>0," * ",""),ifelse((retvaldef$type=="void")&&(retvaldef$reference==0),""," _res ="),ifelse((retvaldef$reference>0)&&(retvaldef$type!="void"),paste("( ",retvaldef$type," * )",sep=""),""),"_obj->",methodname,"(",sep="",file=ccfile,fill=F,append=T)
    lapply(argdefs,function(aa){
                                        # add a checker into R-wrapper to be sure the R-Class of argument is the same as C++-class required by VTK function.
      if(aa$vtkclass || (aa$type=="function")){
        cat("    check.VTK(.",aa$name,",\"",aa$type,"\")",file=Rfile,sep="",fill=T,append=T)
      } 
	#Add protection to callback function against GC
      if(aa$type=="function"){	    
	cat("    assign(\".", methodname ,"\",.",aa$name,",envir=.me)",file=Rfile,sep="",fill=T,append=T)
	cat("    assign(\".", methodname ,".env\", environment(),envir=.me)",file=Rfile,sep="",fill=T,append=T)
       }
    })

                                        # put a .Call to underlying c function in R-file
    cat("    res<-.Call(\"R_",classname,"_",methodname,suffix,"\",",".me$obj",file=Rfile,sep="",fill=F,append=T)
    ccpars <- lapply(argdefs,function(aa){
      if(aa$type=="function"){
        cat(", ","quote(.",aa$name,"(...)), environment()",file=Rfile,sep="",fill=F,append=T)
        argdefs<<-list()
        num.refs<<-0
        return("RVoidFunc, (void *) _arg")
      }
      cat(", ",ifelse(aa$vtkclass,paste(".",aa$name,"$obj",sep=""),paste(".",aa$name,sep="")),file=Rfile,sep="",fill=F,append=T)
      return(paste(ifelse(aa$reference>0,""," * "),"_",aa$name,sep=""))
    })
    cat(paste(ccpars,collapse=", "),file=ccfile,sep="",fill=F,append=T)
#    Pack<-.Pack[[.SubLib]]
    Pack<-""
    cat(",PACKAGE=\"RVTK",Pack,"\")",file=Rfile,sep="",fill=T,append=T)
    cat(");",file=ccfile,sep="",fill=T,append=T)

                                        #create sexp from return value and wrap external pointer to vtk class into R-environment (in case return-value is a pointer to VTK class)

    if(!((retvaldef$type=="void")&&(retvaldef$reference==0))){
      if (retvaldef$vtkclass){
        cat("    res=VtkToSexp(_res,\"",retvaldef$type,"\");",sep="",file=ccfile,fill=T,append=T)
        cat("    res$ret.val <- ",retvaldef$type,"(wrapenv(res$ret.val))",sep="",file=Rfile,fill=T,append=T)
      } else {
        cat("    res=",gsub(" ","_",retvaldef$type),"ToSexp(",ifelse(retvaldef$reference>0,"","&"),"_res,",retvaldef$dim,");",file=ccfile,sep="",fill=T,append=T)
      }
    }

                                        # create return list from the arguments of the function passed as references and wrap external pointer to vtk class into R-environment if needed (in case argument is a pointer to VTK class)

    cat("    PROTECT(ans = allocVector(VECSXP,",num.refs+1 ,"));",file=ccfile,sep="",fill=T,append=T) 
    cat("    PROTECT(res_names = allocVector(VECSXP,",num.refs+1 ,"));",file=ccfile,sep="",fill=T,append=T) 
    cat("    SET_VECTOR_ELT(res_names,0,charToSexp(\"ret.val\",0));",file=ccfile,sep="",fill=T,append=T)
    cat("    SET_VECTOR_ELT(ans,0,res);",file=ccfile,sep="",fill=T,append=T)
    num <- 1
    lapply(argdefs,function(aa){
      if (((aa$weak)||(aa$reference>0))&&(!aa$const)){
        if (aa$vtkclass){
          cat("    SET_VECTOR_ELT(ans,",num,",VtkToSexp(_",aa$name,",\"",aa$type,"\"));\n",sep="",file=ccfile,fill=F,append=T)
          cat("    res$.",aa$name," <- ",aa$type,"(wrapenv(res$.",aa$name,"))",sep="",file=Rfile,fill=T,append=T)
        } else {
          cat("    SET_VECTOR_ELT(ans,",num,",",gsub(" ","_",aa$type),"ToSexp(_",aa$name,",",ifelse(prod(aa$dim)>0,prod(aa$dim),paste("LENGTH(",aa$name,")",sep="")),"));",file=ccfile,sep="",fill=T,append=T)
        }
        cat("    SET_VECTOR_ELT(res_names,",num,",charToSexp(\".",aa$name,"\",0));",file=ccfile,sep="",fill=T,append=T)
        num<<-num+1
      }
    })
    
    cat("    setAttrib(ans, R_NamesSymbol, res_names);",file=ccfile,sep="",fill=T,append=T)
    cat("    UNPROTECT(2);",file=ccfile,sep="",fill=T,append=T)
                                        #complete method definition
    cat("    return ans;",file=ccfile,fill=T,append=T);
    cat("  }",file=ccfile,fill=T,append=T);
    cat(" ",file=ccfile,fill=T,append=T);
                                        #Assign all output from .Call() into .VTK.Cache variable in global environment but return to R only the real output from VTK function. If some data returned as a side effect from function the user can look it in the .VTK.Cache variable just after the function call.
    cat("    assign(\".VTK.Cache\",res,.GlobalEnv)",file=Rfile,fill=T,append=T);
    cat("    return(res$ret.val)",file=Rfile,fill=T,append=T);
    cat("  }",file=Rfile,fill=T,append=T);
    cat(" ",file=Rfile,fill=T,append=T);
    .Method.Names<<-c(methodname,.Method.Names)

  })

                                        #complete class definition
  if(.Has.New){
                                        # New method requires special handling and for all classes it is marked in RHints file with hasnew action. Purely abstract classes does not have New method.
    cat("  SEXP R_",classname,"_New() {",file=ccfile,sep="",fill=T,append=T)
    cat("    SEXP R_res = NULL_USER_OBJECT;",file=ccfile,sep="",fill=T,append=T)
    cat("    ",classname," * res= ( ", classname," * ) ",classname,"::New();",file=ccfile,sep="",fill=T,append=T)
    cat("    R_res=toRPointer(res,\"",classname,"\");",file=ccfile,sep="",fill=T,append=T)
    cat("    return(R_res);",file=ccfile,sep="",fill=T,append=T)
    cat("  }",file=ccfile,sep="",fill=T,append=T)
    cat("  if (identical(.me,environment())){",file=Rfile,sep="",fill=T,append=T)
    cat("   .me$obj <- .Call(\"R_",classname,"_New\")",file=Rfile,sep="",fill=T,append=T)
    #.owner flag does reflects the fact that R owns a copy of VTK object. Delete method of vtkObjectBase checks this flags and decrement the reference counter if the flag is set
#    cat("    .me$.owner <- TRUE",file=Rfile,sep="",fill=T,append=T)
    if (abstract.class){
                                        #The class is abstract according to XML dump but has New initializer. We interpret this as class being a superclass, so New in fact create a different class
      cat("  .me<-wrapvtkclass(.me$obj,.me$GetClassName())",file=Rfile,sep="",fill=T,append=T)
      cat("  reg.finalizer(.me, .me$Delete)",file=Rfile,sep="",fill=T,append=T)
      cat("  return(.me)",file=Rfile,fill=T,append=T)
    } else {
      cat("    reg.finalizer(.me, .me$Delete)",file=Rfile,sep="",fill=T,append=T)
    }
      cat("  }",file=Rfile,sep="",fill=T,append=T)
  }
  cat("#endif",file=ccfile,sep="",fill=T,append=T)
  cat("}",file=ccfile,fill=T,append=T);
  cat("  class(.me) <- c(\"",classname,"\",class(.me))",sep="",file=Rfile,fill=T,append=T)
  cat("  return(.me)",file=Rfile,fill=T,append=T)
  cat("}",file=Rfile,fill=T,append=T);
  cat("}",file=Rfile,fill=T,append=T);
  lapply(.Files,function(x){try(close(x$handle))})
  close(Rfile)
  close(ccfile)
  .Files<<-NULL
}
