#include <R.h>
#include <Rinternals.h>
#include <Rdefines.h>
#include "R_ext/Memory.h"
#include "R_ext/Boolean.h"

#include "Utils.h"
USER_OBJECT_
toRPointer(void *val, const char *type)
{
    USER_OBJECT_ ans;

    PROTECT(ans = R_MakeExternalPtr(val, Rf_install(type), NULL_USER_OBJECT));
    if(type) {
	SET_CLASS(ans, Rf_mkString(type));
    }
    UNPROTECT(1);
  
    return(ans);
}

SEXP voidToSexp(void * x,int unused){
  return(toRPointer(x,"Void Pointer"));
}

SEXP returnNil(){
  return(voidToSexp(NULL,0));
}

void *
getPtrValue(USER_OBJECT_ sval)
{
  if(sval == NULL_USER_OBJECT)
      return(NULL);

  return(R_ExternalPtrAddr(sval));
}

void * SexpTovoid(SEXP z){
  return(getPtrValue(z));
}

void * SexpToVtk(USER_OBJECT_ sval){
  if(sval == NULL_USER_OBJECT)
      return(NULL);

  return(R_ExternalPtrAddr(sval));
}

SEXP VtkToSexp(void *val, const char *type)
{
    SEXP ans;

    PROTECT(ans = R_MakeExternalPtr(val, Rf_install(type), NULL_USER_OBJECT));
    if(type) {
	SET_CLASS(ans, Rf_mkString(type));
    }
    UNPROTECT(1);
  
    return(ans);
}


char *
SexpTochar(USER_OBJECT_ svec)
{
    char * els = CHAR(STRING_ELT(AS_CHARACTER(svec), 0));
    return(els); 
}

SEXP charToSexp(char * val,int unused) {
  SEXP ans=Rf_mkString(val);
  /*  PROTECT(ans = NEW_CHARACTER(1));
  if(val)
      SET_STRING_ELT(ans, 0, COPY_TO_USER_STRING(val));
      UNPROTECT(1);*/
  
  return(ans);
}

SEXP doubleToSexp(double * d, int dims){
  int i;
  SEXP ans;
  double * z;
  if (dims<1) {
    ans=toRPointer(d,"Double Pointer");
  }  else {
    PROTECT(ans = allocVector(REALSXP, dims));
    for (i=0;i<dims;i++){
      REAL(ans)[i]= d[i];
    }
    UNPROTECT(1);
  }
  return ans;
}


double * SexpTodouble(USER_OBJECT_ svec)
{
    double * els = REAL(AS_NUMERIC(svec));
    return(els); 
}

int * SexpToint(USER_OBJECT_ svec)
{
    int * els =  INTEGER(AS_INTEGER(svec));
    return(els); 
}

SEXP intToSexp(int * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
     ans=toRPointer(d,"Int Pointer");
  }  else {
    PROTECT(ans = allocVector(INTSXP, dims));
    for (i=0;i<dims;i++){
      INTEGER(ans)[i]= d[i];
    }
    UNPROTECT(1);
  }
  return ans;
}

short int * SexpToshort_int(USER_OBJECT_ svec)
{
    short int * els = NULL;
 
       int i, n;
 
       n = GET_LENGTH(svec);
       if(n > 0) {
    	els = (short int *) R_alloc(n, sizeof(short int));
    	for(i = 0; i < n; i++) {
    	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
    	}
        }

    return(els); 
}

SEXP short_intToSexp(short int * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
     ans=toRPointer(d,"Int Pointer");
  }  else {
    PROTECT(ans = allocVector(INTSXP, dims));
    for (i=0;i<dims;i++){
      INTEGER(ans)[i]= d[i];
    }
    UNPROTECT(1);
  }
  return ans;
}


SEXP long_unsigned_intToSexp(long unsigned int * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
     ans=toRPointer(d,"Long_Unsigned_Int Pointer");
  } else {  
    PROTECT(ans = allocVector(INTSXP, dims));
    for (i=0;i<dims;i++){
      INTEGER(ans)[i]= d[i];
    }
    UNPROTECT(1);
  }
  return ans;
}

SEXP long_intToSexp(long int * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
     ans=toRPointer(d,"Long_Int Pointer");
  }  else {
    PROTECT(ans = allocVector(INTSXP, dims));
    for (i=0;i<dims;i++){
      INTEGER(ans)[i]= d[i];
    }
  
  UNPROTECT(1);
  }
  return ans;
}

SEXP long_long_intToSexp(long long int * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
     ans=toRPointer(d,"Long_long_Int Pointer");
  }  else {
    PROTECT(ans = allocVector(INTSXP, dims));
    for (i=0;i<dims;i++){
      INTEGER(ans)[i]= d[i];
    }
  
  UNPROTECT(1);
  }
  return ans;
}

SEXP long_long_unsigned_intToSexp(long long unsigned int * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
     ans=toRPointer(d,"Long_long_unsigned_Int Pointer");
  }  else {
    PROTECT(ans = allocVector(INTSXP, dims));
    for (i=0;i<dims;i++){
      INTEGER(ans)[i]= d[i];
    }
  
  UNPROTECT(1);
  }
  return ans;
}


SEXP floatToSexp(float * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
    ans=toRPointer(d,"Float Pointer");
  }  else {
    PROTECT(ans = allocVector(REALSXP, dims));
    for (i=0;i<dims;i++){
      REAL(ans)[i]= d[i];
    }
    UNPROTECT(1);
  }
  return ans;
}


float * SexpTofloat(USER_OBJECT_ svec)
{
    float * els = NULL;
 
    int i, n;
 
    n = GET_LENGTH(svec);
    if(n > 0) {
	els = (float *) R_alloc(n, sizeof(float));
	for(i = 0; i < n; i++) {
	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
	}
    }

    return(els); 
}

long int * SexpTolong_int(USER_OBJECT_ svec)
{
    long int * els = NULL;
 
    int i, n;
 
    n = GET_LENGTH(svec);
    if(n > 0) {
	els = (int *) R_alloc(n, sizeof(long int));
	for(i = 0; i < n; i++) {
	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
	}
    }

    return(els); 
}

long long int * SexpTolong_long_int(USER_OBJECT_ svec)
{
    long long int * els = NULL;
 
    int i, n;
 
    n = GET_LENGTH(svec);
    if(n > 0) {
	els = (long long int *) R_alloc(n, sizeof(long long int));
	for(i = 0; i < n; i++) {
	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
	}
    }

    return(els); 
}

long long unsigned int * SexpTolong_long_unsigned_int(USER_OBJECT_ svec)
{
    long long unsigned int * els = NULL;
 
    int i, n;
 
    n = GET_LENGTH(svec);
    if(n > 0) {
	els = (long long unsigned int *) R_alloc(n, sizeof(long long unsigned int));
	for(i = 0; i < n; i++) {
	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
	}
    }

    return(els); 
}

unsigned char * SexpTounsigned_char(USER_OBJECT_ svec)
{
    unsigned char  * els = NULL;
 
    int i, n;
 
    n = GET_LENGTH(svec);
    if(n > 0) {
	els = (int *) R_alloc(n, sizeof(unsigned char));
	for(i = 0; i < n; i++) {
	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
	}
    }

    return(els); 
}

signed char * SexpTosigned_char(USER_OBJECT_ svec)
{
    signed char  * els = NULL;
 
    int i, n;
 
    n = GET_LENGTH(svec);
    if(n > 0) {
	els = (int *) R_alloc(n, sizeof(signed char));
	for(i = 0; i < n; i++) {
	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
	}
    }

    return(els); 
}

long unsigned int * SexpTolong_unsigned_int(USER_OBJECT_ svec)
{
    long unsigned int  * els = NULL;
 
    int i, n;
 
    n = GET_LENGTH(svec);
    if(n > 0) {
	els = (long unsigned int *) R_alloc(n, sizeof(long unsigned int));
	for(i = 0; i < n; i++) {
	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
	}
    }

    return(els); 
}

unsigned int * SexpTounsigned_int(USER_OBJECT_ svec)
{
    unsigned int  * els = NULL;
 
    int i, n;
 
    n = GET_LENGTH(svec);
    if(n > 0) {
	els = (unsigned int *) R_alloc(n, sizeof(unsigned int));
	for(i = 0; i < n; i++) {
	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
	}
    }

    return(els); 
}

short unsigned int * SexpToshort_unsigned_int(USER_OBJECT_ svec)
{
    short unsigned int  * els = NULL;
 
    int i, n;
 
    n = GET_LENGTH(svec);
    if(n > 0) {
	els = (short unsigned int *) R_alloc(n, sizeof(short unsigned int));
	for(i = 0; i < n; i++) {
	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
	}
    }

    return(els); 
}

SEXP unsigned_charToSexp(unsigned char * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
     ans=toRPointer(d,"unsigned_char Pointer");
  } else {  
    PROTECT(ans = allocVector(INTSXP, dims));
    for (i=0;i<dims;i++){
      INTEGER(ans)[i]= d[i];
    }
    UNPROTECT(1);
  }
  return ans;
}

SEXP signed_charToSexp(signed char * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
     ans=toRPointer(d,"unsigned_char Pointer");
  } else {  
    PROTECT(ans = allocVector(INTSXP, dims));
    for (i=0;i<dims;i++){
      INTEGER(ans)[i]= d[i];
    }
    UNPROTECT(1);
  }
  return ans;
}

SEXP short_unsigned_intToSexp(short unsigned int * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
     ans=toRPointer(d,"short_unsigned_int Pointer");
  } else {  
    PROTECT(ans = allocVector(INTSXP, dims));
    for (i=0;i<dims;i++){
      INTEGER(ans)[i]= d[i];
    }
    UNPROTECT(1);
  }
  return ans;
}

SEXP unsigned_intToSexp(unsigned int * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
     ans=toRPointer(d,"unsigned_int Pointer");
  } else {  
    PROTECT(ans = allocVector(INTSXP, dims));
    for (i=0;i<dims;i++){
      INTEGER(ans)[i]= d[i];
    }
    UNPROTECT(1);
  }
  return ans;
}




void RVoidFunc(void * arg){
  RVoidFuncArg * z = arg;
  eval(z->func, z->env);
}

void RVoidFuncArgDelete(void * arg){
  Free(arg);
}
